import base64
import uuid
from urllib.parse import SplitResult, ParseResult

from Apifon.sdk.Resource.Helpers.OAuthGrantType import OAuthGrantType
from Apifon.sdk.Resource.Helpers.OpenIDScope import OpenIDScope
from Apifon.sdk.Resource.Helpers.S256PKCEChallenge import S256PKCEChallenge


class OAuthConfigurationBuilder:
    def __init__(self, client_id, grant_type):
        self.grant_type = grant_type
        self.client_id = client_id
        self.scope = set()
        self.client_secret = None
        self.pkce_challenge = None
        self.redirect_uri = None
        self.state = None

    @staticmethod
    def get_client_credentials_builder(client_id, client_secret):
        configBuilder = OAuthConfigurationBuilder(client_id, OAuthGrantType.CLIENT_CREDENTIALS)
        configBuilder.client_secret = client_secret
        #configBuilder.scope.add(OpenIDScope.openid.value)
        #configBuilder.scope.add(OpenIDScope.accountInfo.value)
        return configBuilder

    def with_scope(self, scope):
        if scope is None or not OpenIDScope.has_value(scope):
            raise TypeError("Invalid scope. Should be of type OpenIDScope.value")
        self.scope.add(scope)
        return self

    def with_scopes(self, scopes):
        if scopes is None or not isinstance(scopes, set):
            raise TypeError("Invalid scopes. Should be of type set")
        for scope in scopes:
            try:
                self.with_scope(scope)
            except TypeError:
                raise
        return self

    def with_redirect_uri(self, redirect_uri, forceUrlEncoding=True):
        if redirect_uri is None or not isinstance(redirect_uri, str) and not isinstance(redirect_uri,
                                                                                        SplitResult) and not isinstance(
                redirect_uri, ParseResult):
            raise TypeError("Invalid redirect uri. Should be of type str, SplitResult or ParseResult")
        if isinstance(redirect_uri, str):
            self.redirect_uri = redirect_uri
        elif isinstance(redirect_uri, SplitResult) or isinstance(redirect_uri, ParseResult):
            self.redirect_uri = redirect_uri.geturl()
        return self

    def with_state(self, state):
        if state is None or not isinstance(state, str):
            raise TypeError("Invalid state. Should be of type str")
        self.state = state
        return self

    def build(self):
        return OAuthConfigurationBuilder.OAuthConfiguration(self)

    class OAuthConfiguration:
        def __init__(self, OAuthConfigurationBuilderInstance):
            self.grant_type = None
            self.client_id = None
            self.scope = list()
            self.client_secret = None
            self.pkce_challenge = None
            self.redirect_uri = None
            self.state = None
            self.set_values(OAuthConfigurationBuilderInstance)

        def set_values(self, OAuthConfigurationBuilderInstance):
            self.grant_type = OAuthConfigurationBuilderInstance.grant_type
            self.client_id = OAuthConfigurationBuilderInstance.client_id
            self.scope = list(OAuthConfigurationBuilderInstance.scope)
            self.client_secret = OAuthConfigurationBuilderInstance.client_secret
            if OAuthConfigurationBuilderInstance.pkce_challenge is not None:
                self.pkce_challenge = S256PKCEChallenge(OAuthConfigurationBuilderInstance.pkce_challenge.code_verifier)
            else:
                self.pkce_challenge = None
            self.redirect_uri = OAuthConfigurationBuilderInstance.redirect_uri
            if OAuthConfigurationBuilderInstance.state is None:
                self.state = base64.b64encode(str(uuid.uuid4()).encode("utf-8")).decode()
            else:
                self.state = OAuthConfigurationBuilderInstance.state

        def get_scopes_as_string(self):
            scopeList = list()
            for scope in self.scope:
                scopeList.append(str(OpenIDScope(scope).name))

            scopeString = '+'.join(scopeList)
            return scopeString

