import http.client
import json
import sys
import traceback
from io import StringIO
from datetime import datetime, timezone

from Apifon.sdk.Exception.ApifonRestError import ApifonRestError
from Apifon.sdk.Exception.OAuthError import OAuthError
from Apifon.sdk.Exception.OAuthTokenError import OAuthTokenError
from Apifon.sdk.OAuthRegistry import OAuthRegistry
from Apifon.sdk.Response.ApiErrorResponse import ApiErrorResponse
from Apifon.sdk.Security.Signature import Signature


class Mookee:
    __instance = None
    active_oauth_identifier = None
    auto_refresh_tokens = False
    production_url = "ars.apifon.com"
    staging_url = "ars.staging.apifon.com"
    identity_production_server_url = "https://ids.apifon.com"
    identity_staging_server_url = "https://ids-staging.apifon.com"

    def __init__(self):

        self.base_url = None
        self.credentials = {}
        self.staging = False
        self.key = ""

    @staticmethod
    def get_instance():
        if Mookee.__instance is None:
            Mookee.__instance = Mookee()

        return Mookee.__instance
    '''
     * Sets the currently active OAuth client identifier.<br><br>
     * <b>Note</b>: Setting OAuth mode will force Mookee to dispatch <b>ALL</b>
     * requests using the given OAuth client credentials for this identifier.
     *
    '''

    @staticmethod
    def set_mookee_to_oauth_mode(identifier, auto_refresh_tokens):
        if Mookee.is_oauth_enabled(identifier):
            Mookee.active_oauth_identifier = identifier
            Mookee.auto_refresh_tokens = auto_refresh_tokens
            return
        raise ValueError("Identifier " + identifier + " is not a valid OAuth identifier.")

    '''
     * Sets mookee to the HMAC token/secret key mode.<br><br>
     * <b>Note</b>: Setting mookee on HMAC token mode will force Mookee to use
     * Token credentials for <b>ALL</b> requests.<br>
     * This will <b>NOT</b> invalidate any active OAuth2 token credentials already registered.
    '''

    @staticmethod
    def set_mookee_to_hmac_mode():
        Mookee.active_oauth_identifier = None

    @staticmethod
    def is_oauth_enabled(identifier):
        return OAuthRegistry.INSTANCE.is_oauth_enabled(identifier)

    '''
    * Set enable to True if you want to use our Staging Server.
    '''

    @staticmethod
    def set_staging(enable=False, host=None, identity_server_host=None, ):
        Mookee.__instance.staging = enable
        if host is not None:
            Mookee.staging_url = host
        if identity_server_host is not None:
            Mookee.identity_staging_server_url = identity_server_host


    '''
     * Adds a given OAuth configuration for the given identifier.<br>
     * This will overwrite any existing configuration AND revoke any (if any)
     * existing tokens for this identifier.
    '''
    @staticmethod
    def add_oauth_resource(identifier, oauth_cofiguration):
        try:
            OAuthRegistry.INSTANCE.add_oauth_resource(identifier, oauth_cofiguration)
        except OAuthError:
            raise

    '''
     * Removes the OAuthResource with the given identifier. This will also
     * revoke any active tokens and refresh tokens (if any) for the given
     * identifier.
    '''
    @staticmethod
    def remove_oauth_resource(identifier, revoke_tokens):
        try:
            OAuthRegistry.INSTANCE.remove_oauth_resource(identifier, revoke_tokens)
        except OAuthError:
            raise

    '''
     * Tries to revoke tokens (access token and refresh token if available) for
     * the given identifier.<br><br>
     * Note that after this method completes successfully your tokens will be
     * invalid and any authorization must be performed again.<br>
     * This will NOT remove the OAuthResource from the Registry, you can call
     * {@link #removeOAuthResource() for that.
     *
    '''
    @staticmethod
    def revoke_tokens(identifier):
        try:
            OAuthRegistry.INSTANCE.revoke_tokens(identifier)
        except OAuthError:
            raise

    '''
     * Tries to retrieve access token for the client credentials OAuth grant.
    '''
    @staticmethod
    def retrieve_client_credentials_token(identifier):
        try:
            OAuthRegistry.INSTANCE.retrieve_tokens_for_client_credential(identifier)
        except OAuthError:
            raise

    '''     
     * Tries to retrieve access token for the client credentials OAuth grant and
     * returns a copy of the tokens.
     '''

    @staticmethod
    def retrieve_client_credentials_and_get_token(identifier):
        try:
            return OAuthRegistry.INSTANCE.retrieve_tokens_for_client_credential_and_get(identifier)
        except OAuthError:
            raise

    '''
    * Refresh active access token. This will revoke any active tokens and issue
    '''
    @staticmethod
    def refresh_access_token(identifier):
        try:
            return OAuthRegistry.INSTANCE.refresh_tokens(identifier)
        except OAuthError:
            raise

    # private set date with timezone
    @staticmethod
    def get_request_date():
        return datetime.now(timezone.utc).strftime('%a, %d %b %Y %H:%M:%S GMT')

    # Add credentials
    @staticmethod
    def add_credentials(key, token, private_key):
        l = list()
        l.append(token)
        l.append(private_key)
        Mookee.__instance.credentials[key] = l

    # Set Active Credential
    @staticmethod
    def set_active_credential(key):
        Mookee.__instance.key = key

    @staticmethod
    def dispatch(resource):
        if Mookee.__instance.staging:
            Mookee.__instance.base_url = Mookee.__instance.staging_url
        else:
            Mookee.__instance.base_url = Mookee.__instance.production_url

        endpoint = resource.cur_endpoint
        body = resource.cur_body
        method = resource.cur_method
        try:
            return Mookee.__instance.make_call(endpoint, method, body)
        except PermissionError:
            print("401 Unauthorized")
        except Exception as x:
            print(x)


    @staticmethod
    def dispatch_to_response(resource):
        if Mookee.__instance.staging:
            Mookee.__instance.base_url = Mookee.__instance.staging_url
        else:
            Mookee.__instance.base_url = Mookee.__instance.production_url

        endpoint = resource.cur_endpoint
        body = resource.cur_body
        method = resource.cur_method
        headers = resource.cur_headers
        try:
            if Mookee.active_oauth_identifier is not None:
                response = Mookee.__instance.make_oauth_call_to_response(endpoint, method, body, headers)
            else:
                response = Mookee.__instance.make_call_to_response(endpoint, method, body, headers)

            if response.status >= 400:
                if response.status == 401:
                    raise PermissionError
                try:
                    current_sys_stdout = sys.stdout
                    sys.stdout = buffer = StringIO()
                    print(response.read().decode('utf8'))
                    badRespStr = buffer.getvalue()
                    sys.stdout = current_sys_stdout
                    badResp = json.loads(badRespStr)
                except Exception:
                    exceptionBundle = {"status": response.status, "title": response.reason}
                    raise ApifonRestError(ApiErrorResponse(exceptionBundle))
                raise ApifonRestError(ApiErrorResponse(badResp))
            else:
                return response
        except ApifonRestError:
            raise
        except PermissionError:
            print("401 Unauthorised")
            traceback.print_exc()
        except Exception as x:
            print("Exception while dispatching request ", x)
            traceback.print_exc()

    @staticmethod
    def make_call(endpoint, method, body):

        if Mookee.active_oauth_identifier is not None:
            response = Mookee.__instance.make_oauth_call_to_response(endpoint, method, body, {})
        else:
            response = Mookee.__instance.make_call_to_response(endpoint, method, body, {})
        if response.status == 401:
            raise PermissionError
        t = json.loads(response.read().decode('utf8'))
        return t


    @staticmethod
    def make_call_to_response(endpoint, method, body, headers):
        if body is None:
            body = ""

        # Get the date from internal function
        request_date = Mookee.__instance.get_request_date()

        # Build the message
        signed_message = method + "\n" + endpoint.split('?')[0] + "\n" + body + "\n" + request_date

        # Get the signature from the sign class
        signature = Signature.sign(signed_message, Mookee.__instance.credentials[Mookee.__instance.key][1])

        # Build the authorization header
        headers.update({
            "Content-type": "application/json",
            "Authorization": "ApifonWS %s: %s" % (Mookee.__instance.credentials[Mookee.__instance.key][0], signature),
            "X-ApifonWS-Date": request_date
        })

        # Make the request call
        connection = http.client.HTTPConnection(Mookee.__instance.base_url)
        connection.request(method, endpoint, body, headers)
        t = connection.getresponse()
        return t

    @staticmethod
    def make_oauth_call_to_response(endpoint,method, body, headers):
        if body is None:
            body = ""

        # Build the authorization header
        headers.update({
            "Authorization": "Bearer " + str(Mookee.get_active_oauth_credential()),
            "Content-type": "application/json"
        })

        # Make the request call
        try:
            connection = http.client.HTTPConnection(Mookee.__instance.base_url)
            connection.request(method, endpoint, body, headers)
        except Exception as e:
            raise
        t = connection.getresponse()
        if t.status == 401:
            if Mookee.auto_refresh_tokens:
                OAuthRegistry.INSTANCE.refresh_tokens(Mookee.active_oauth_identifier)
                headers.update({ "Authorization": "Bearer " + str(Mookee.get_active_oauth_credential())})
                connection = http.client.HTTPConnection(Mookee.__instance.base_url)
                connection.request(method, endpoint, body, headers)
                newT = connection.getresponse()
                if newT.status == 401:
                    raise OAuthTokenError("Status: 401\nError:" + newT.getheader("WWW-Authenticate"))
                return newT

            else:
                raise OAuthError("Unauthorized to access API")
        else:
            return t



    @staticmethod
    def get_active_oauth_credential():
        oauth_credentials = OAuthRegistry.INSTANCE.retreive_active_tokens(Mookee.active_oauth_identifier)
        if oauth_credentials.has_expired() and Mookee.auto_refresh_tokens:
            OAuthRegistry.INSTANCE.refresh_tokens(Mookee.active_oauth_identifier)
            return OAuthRegistry.INSTANCE.retreive_active_tokens(Mookee.active_oauth_identifier).access_token
        return oauth_credentials.access_token


    def get_identity_server_url(self):
        if self.staging_url is not None and self.staging:
            return self.identity_staging_server_url
        return self.identity_production_server_url
