import copy
from enum import Enum

from Apifon.sdk.Exception.OAuthError import OAuthError
from Apifon.sdk.Model.Request.OAuthConfigurationBuilder import OAuthConfigurationBuilder
from Apifon.sdk.Resource.OAuthResource import OAuthResource


class OAuthRegistry(Enum):
    INSTANCE = "instance"

    def __init__(self, oauth_registry={}):
        self.oauth_registry = {}

    '''
     * Tries to add an OAuthResource with the given identifier.
     * This will revoke and remove any existing OAuthResource and
     * OAuthCredentials for this identifier.
    '''

    def add_oauth_resource(self, identifier, oauth_configuration):
        try:
            if oauth_configuration is None or not isinstance(oauth_configuration,
                                                             OAuthConfigurationBuilder.OAuthConfiguration):
                raise TypeError("Invalid oauth configuration. Should be of type OAuthConfiguration")
            oauth_resource = OAuthResource(oauth_configuration)

            if identifier in self.oauth_registry:
                self.remove_oauth_resource(identifier, True)

            self.oauth_registry[identifier] = oauth_resource
        except OAuthError:
            raise

    '''
     * Removes the given OAuth info related to the given identifier from the
     * registry and revokes any tokens (if any)
     * No action will be performed if the identifier is invalid.
    '''

    def remove_oauth_resource(self, identifier, revoke_tokens):
        try:
            if identifier in self.oauth_registry:
                if revoke_tokens:
                    self.revoke_tokens(identifier)
                self.oauth_registry.pop(identifier, None)
        except OAuthError:
            raise

    '''
     * Tries to revoke tokens (access token and refresh token if available) for
     * the given resource.
     * Note that after this method completes successfully your tokens will be
     * invalid and any authorization must be performed again.
    '''

    def revoke_tokens(self, identifier):
        try:
            self.validate_oauth_enabled(identifier)
            self.oauth_registry[identifier].revoke_tokens()
        except OAuthError:
            raise

    def is_oauth_enabled(self, identifier):
        return identifier in self.oauth_registry

    def get_resource_for_identifier(self, identifier):
        try:
            self.validate_oauth_enabled(identifier)
            return self.oauth_registry[identifier]
        except OAuthError:
            raise

    def retrieve_tokens_for_client_credential(self, identifier):
        try:
            self.validate_oauth_enabled(identifier)
            resource = self.oauth_registry[identifier]
            resource.get_tokens_for_client_credentials_grant()
        except OAuthError:
            raise

    def retrieve_tokens_for_client_credential_and_get(self, identifier):
        try:
            self.validate_oauth_enabled(identifier)
            resource = self.oauth_registry[identifier]
            resource.get_tokens_for_client_credentials_grant()
            return copy.deepcopy(resource.oauth_credentials)
        except OAuthError:
            raise

    def refresh_tokens(self, identifier):
        try:
            self.validate_oauth_enabled(identifier)
            self.oauth_registry[identifier].refresh_token()
        except OAuthError:
            raise

    def retreive_active_tokens(self, identifier):
        if self.is_oauth_enabled(identifier):
            return copy.deepcopy(self.get_resource_for_identifier(identifier).oauth_credentials)
        raise OAuthError("Active tokens for " + identifier + " not located.")

    def validate_oauth_enabled(self, identifier):
        if not self.is_oauth_enabled(identifier):
            raise OAuthError("Identifier " + identifier + " doesn't have OAuth enabled.")
